;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; jq wrapper for processing S-expressions.
;;;
;;; Copyright (c) 2020, Evan Hanson
;;;
;;; See LICENSE for details.
;;;

(declare (module sq))

(import (chicken condition)
        (chicken file posix)
        (chicken format)
        (chicken foreign)
        (chicken io)
        (chicken irregex)
        (chicken port)
        (chicken pretty-print)
        (chicken process)
        (chicken process-context)
        (chicken string)
        (chicken type)
        (optimism getopt)
        (only (simple-exceptions) message arguments)
        (srfi 18)
        (sq srfi 180))

(import-syntax (begin-syntax)
               (matchable)
               (only (miscmacros) inc!))

(import-for-syntax (chicken io)
                   (chicken process))

(define-syntax begin0-let*
  (syntax-rules ()
    ((_ ((x y) . rest) . body)
     (let* ((x y) . rest) (begin . body) x))))

(define-foreign-variable C_abort_on_thread_exceptions bool)

(define options
  '(((-j --json))
    ((-h --help))
    ((--version))
    ((--debug))))

(define help "Usage: sq [-j] [-- <jq-args>] [<jq-expr>]")

(define version
  (begin-syntax
    (let ((x (with-input-from-pipe "git describe --tags 2>/dev/null" read-line)))
      (if (string? x) x "(egg)"))))

(define exit-status
  (make-parameter 0))

(: make-concatenated-port* (#!rest (or string input-port) -> input-port))
(define (make-concatenated-port* . args)
  (let ((ports (map (lambda (x) (if (string? x) (open-input-string x) x)) args)))
    (apply make-concatenated-port ports)))

(define (make-multiplexed-input-port output-port #!optional (port (current-input-port)))
  (make-input-port
   (lambda ()
     (begin0-let* ((x (read-char port)))
       (when (char? x) (write-char x output-port))))
   (lambda () (char-ready? port))
   (lambda () (close-input-port port))
   (lambda () (peek-char port))
   (lambda (_ n s i)
     (begin0-let* ((x (read-string! n s port i)))
       (write-string (substring s i (+ i x)) x output-port)))
   (lambda (_ n)
     (begin0-let* ((x (read-line port n)))
       (when (string? x) (write-line x output-port))))))

(define (make-location-tracking-input-port #!optional (port (current-input-port)))
  (let ((line 1)
        (column 0))
    (values
     (make-input-port
      (lambda ()
        (begin0-let* ((x (read-char port)))
          (cond ((eof-object? x))
                ((eq? x #\newline)
                 (set! column 0)
                 (inc! line))
                (else
                 (inc! column)))))
      (lambda () (char-ready? port))
      (lambda () (close-input-port port))
      (lambda () (peek-char port)))
     (lambda () (values line column)))))

(define (make-rewindable-input-port #!optional (port (current-input-port)))
  (let ((buffer (open-output-string)))
    (values (make-multiplexed-input-port buffer port)
            (lambda () (make-concatenated-port* (get-output-string buffer) port)))))

(define (process-start! thunk)
  (let*-values (((c-i p-o) (create-pipe))
                ((p-i c-o) (create-pipe))
                ((pid)     (process-fork
                            (lambda ()
                              (file-close p-i)
                              (file-close p-o)
                              (duplicate-fileno c-i fileno/stdin)
                              (duplicate-fileno c-o fileno/stdout)
                              (thunk)))))
    (file-close c-i)
    (file-close c-o)
    (values pid (open-input-file* p-i) (open-output-file* p-o))))

(define (thread-waiter port #!optional (mode #:all))
  (condition-case (let ((fd (port->fileno port)))
                    (lambda () (thread-wait-for-i/o! fd #:input)))
    ((exn type) void)))

(define (copy-port* input output #!optional (reader read-char) (writer write-char))
  (let ((thread-wait-for-input! (thread-waiter input #:input)))
    (copy-port
     input
     output
     (lambda (p)
       (thread-wait-for-input!)
       (reader p))
     (lambda (x p)
       (writer x p)))))

(define (proxy input output)
  (let-values (((input* rewind) (make-rewindable-input-port input)))
    (handle-exceptions e
        (if ((condition-predicate 'syntax) e)
            (proxy-string (rewind) output)
            (proxy-scheme (rewind) output))
      ;; skip whitespace
      (read-token char-whitespace? input*)
      ;; look for a parenthesis or valid scheme datum
      (or (eq? (peek-char input*) #\()
          (read input*))
      (signal 'scheme))))

(define (proxy-scheme input output)
  (copy-port* input output read json-write))

(define (proxy-json input output)
  (copy-port* input output json-read pretty-print))

(define (proxy-string input output)
  (copy-port*
   input
   output
   (lambda (p) (read-string 65535 p))
   (lambda (x p) (write-string x 65535 p))))

(define syntax-error?
  (condition-predicate 'syntax))

(define (->string* x)
  (with-output-to-string (lambda () (write x))))

(define (error-message e location)
  (format "~a: ~a~a~n"
          (if (json-error? e) "parse error" "error")
          (if (or (json-error? e) (syntax-error? e))
              (let-values (((line column) (location)))
                (irregex-replace
                 "[.!]?$"
                 (if (json-error? e) (json-error-reason e) (message e))
                 (format " in form ending at line ~a, column ~a" line column)))
              (message e))
          (if (or (json-error? e) (null? (arguments e)))
              (string)
              (string-append ": " (string-intersperse (map ->string* (arguments e)) ", ")))))

(define (process-input args opts)
  (let-values (((port location) (make-location-tracking-input-port)))
    (handle-exceptions e
        (begin
          (display (error-message e location) (current-error-port))
          (when (assq '--debug opts) (signal e))
          (exit (max (exit-status) 1)))
      (let*-values (((pid input output)
                     (process-start!
                      (lambda ()
                        (condition-case (process-execute "jq" args)
                          ((exn process) (exit 127))
                          (() (exit 1))))))
                    ((writer)
                     (thread-start!
                      (lambda ()
                        (proxy port output))))
                    ((reader)
                     (thread-start!
                      (lambda ()
                        (if (or (assq '-j opts)
                                (assq '--json opts)
                                (member "-r" args)
                                (member "--raw-output" args))
                            (proxy-string input (current-output-port))
                            (proxy-json input (current-output-port)))))))
        (handle-exceptions e
            (begin
              (close-output-port output)
              (close-input-port input)
              (receive (pid* ok status) (process-wait pid)
                (unless (zero? status) (exit-status status))
                (unless (eq? e 'ok) (signal e))))
          (thread-join! writer)
          (close-output-port output)
          (thread-join! reader)
          (close-input-port input)
          (signal 'ok))))))

(define (main args)
  (let ((opts (parse-command-line args options)))
    (when (or (assq '-h opts) (assq '--help opts))
      (print help)
      (exit))
    (when (or (assq '-v opts) (assq '--version opts))
      (print version)
      (exit))
    (parameterize ((parentheses-synonyms #f))
      (process-input (alist-ref '-- opts) opts)
      (exit (exit-status)))))

(cond-expand
  (compiling
   (set! C_abort_on_thread_exceptions #t)
   (main (command-line-arguments)))
  (else))
